/**
 * @file	ssd1309.h
 * @author	Robert Steigemann
 * @date	15 January 2017
 * @brief	SSD1309 OLED Driver library
 *
 * This library provides high level functions to use an OLED Display with SSD1309 controller.
 * You have to configure the interface and pins in this header file.
 *
 * @license creative commons license CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/
 */

#ifndef SSD1309_H_
#define SSD1309_H_

/*#############################################################################
########################## select your interface here #########################
#############################################################################*/

//#define SSD1309_68XX		// 68XX 8-Bit interface, not implemented yet
//#define SSD1309_80XX		// 80XX 8-Bit interface, not implemented yet
#define	SSD1309_SPI			// SPI (5-Wire) interface


/*#############################################################################
######################### config 68XX here, if chosen #########################
#############################################################################*/

#ifdef SSD1309_68XX

// TODO

#endif


/*#############################################################################
######################## config 80XX here, if chosen ##########################
#############################################################################*/

#ifdef SSD1309_80XX

// TODO

#endif


/*#############################################################################
######################### config SPI here, if chosen ##########################
#############################################################################*/

#ifdef SSD1309_SPI


#define SSD1309_PORT	PORTD_OUT	// The used PORT
#define SSD1309_DIR		PORTD_DIR	// The corresponding DDR Register

#define SSD1309_SPI_I	SPID		// the used SPI port

#define SSD1309_CS		2			// Chip select (GPIO)
#define SSD1309_DC		4			// Data/Command (GPIO)
#define SSD1309_RES		3			// Reset (GPIO)
#define SSD1309_SCK		7			// Serial Clock		
#define SSD1309_MOSI	5			// Master Out Display In


#endif


/*#############################################################################
############################# select display size #############################
#############################################################################*/

#define SSD1309_COL		128			
#define SSD1309_ROW		64


/*#############################################################################
############################# function prototypes #############################
#############################################################################*/

/**
  * @brief  initialize the display
  *			This function will configure the SPI-Interface and initialize and clean the display.
  *			Have to be executed before any other display function.
  *
  * @param	rotate		rotates the complete display by 180 degrees
  * @param  contrast    specifies the contrast level at init
  */
void ssd1309_init(bool rotate, uint8_t contrast);

/**
  * @brief  clears the whole display by writing 0x00 to memory
  */
void ssd1309_clear(void);

/**
  * @brief  set brightness of the display
  *			This function will set the pixel output current of the OLED-Display. 
  *
  * @param  contrast    specifies the contrast level (0...255)
  */
void ssd1309_contrast(uint8_t contrast);

/**
  * @brief  prints a string at specified position
  *
  * @param	*aString	pointer to the string to print (0 terminated)
  * @param  page		the page (line) where the string is displayed
  * @param  startColumn	the column where the string starts
  */
void ssd1309_print(char *aString, uint8_t page, uint8_t startColumn);

/**
  * @brief  prints a string containing a number at specified position
  *
  * Only the following chars and whitespaces are allowed: 0123456789,./+-
  *
  * @param	*aString	pointer to the string to print (0 terminated)
  * @param  page		the page (line) where the string is displayed
  * @param  startColumn	the column where the string starts
  */
void ssd1309_print_bigDigit(char *aString, uint8_t page, uint8_t startColumn);

/**
  * @brief  prints a string at specified position
  *
  * @param	aString	pointer to the PROGMEM string to print (0x00 terminated)
  * @param  page		the page (line) where the string is displayed
  * @param  startColumn	the column where the string starts
  */
void ssd1309_print_P(PGM_P aString, uint8_t page, uint8_t startColumn);

/**
  * @brief  prints a formatted string at specified position
  *
  * @param  page		the page (line) where the string is displayed
  * @param  startColumn	the column where the string starts
  */
void ssd1309_printf(uint8_t page, uint8_t startColumn, const char* __fmt, ...);

/**
  * @brief  prints a formatted string at specified position, format string in program memory
  *
  * @param  page		the page (line) where the string is displayed
  * @param  startColumn	the column where the string starts
  */
void ssd1309_printf_P(uint8_t page, uint8_t startColumn, PGM_P __fmt, ...);

/**
  * @brief  prints a formatted string with a number in big digits at specified position
  *
  * Only the following chars and whitespaces are allowed: 0123456789,./+-
  *
  * @param  page		the page (line) where the string is displayed
  * @param  startColumn	the column where the string starts
  */
void ssd1309_printf_bigDigit(uint8_t page, uint8_t startColumn, const char* __fmt, ...);

/**
  * @brief  prints a formatted string with a number in big digits at specified position, format string in program memory
  *
  * Only the following chars and whitespaces are allowed: 0123456789,./+-
  *
  * @param  page		the page (line) where the string is displayed
  * @param  startColumn	the column where the string starts
  */
void ssd1309_printf_bigDigit_P(uint8_t page, uint8_t startColumn, PGM_P __fmt, ...);

/**
  * @brief  shows a picture at specified position
  *         Use the LCD Assistant tool from http://en.radzio.dxp.pl/bitmap_converter/ to generate
  *         pattern from an image. Select vertical byte orientation, little endianness and 8 pixels/byte.
  *
  * @param	*pic		pointer to the picture data generated by the Image2GLCD tool, type have to be a "const uint8_t PROGMEM" array
  * @param  startPage	the page where the picture starts, each page is 8 pixels high
  * @param  endPage		the page where the picture ends, height of the picture have to be a multiple of 8 and >= startPage
  * @param  startCol	x offset for display position 
  * @param  totalCol	equals with of the picture
  */
void ssd1309_picture(const uint8_t *pic, uint8_t startPage, uint8_t endPage, uint8_t startCol, uint8_t totalCol);

/**
  * @brief  draws a bar graph at specified position
  *
  * @param	percent		the filling of the graph (0-100%)
  * @param  startPage	the page where the bar graph starts, each page is 8 pixels high
  * @param  endPage		the page where the bar graph ends, height of the bar graph have to be a multiple of 8 and >= startPage
  * @param  startCol	x offset for display position 
  * @param  totalCol	equals with of the bar graph, have to be >=5
  */
void ssd1309_bargraph(uint8_t percent, uint8_t startPage, uint8_t endPage, uint8_t startCol, uint8_t totalCol);

#endif /* SSD1309_H_ */